#ifndef MODEL_HPP
#define MODEL_HPP

#include <cmath>
#include <iostream>
#include <valarray>
#include <vector>
using namespace std;

class ODE{
    protected:
    int n;
    valarray<double> (*flow)(valarray<double> a);
    vector< valarray<double> > Xn; // Justifier Valarray
    public:
    ODE(){
        n=0;
        flow=NULL;
    }
    ODE(int m, valarray<double> X0){
        n=m;
        flow=NULL;
        Xn.push_back(X0);
    }
    ODE(const ODE & o){
        n = o.n;
        flow = o.flow;
        Xn = o.Xn; 
    }
    // est ce necessaire de reimplementer le constructeur par copie et destructeur , justifier ?
    void set_flow( valarray<double> (*p)(valarray<double> a));
    double operator [](int);
    virtual valarray<double>compute_flux(valarray<double> x);
    virtual void solve_dt(double );

};

class Poisson_system: public ODE{
    protected:
    double (*Ham)(valarray<double> a);
    valarray<double> (*GradH)(valarray<double> a);
    valarray<double> (*Jx)(valarray<double> a,valarray<double> b);
   

    public:
        Poisson_system(): ODE(){}; // justify un peu
        Poisson_system(int m, valarray<double> X0): ODE(m,X0){};
        Poisson_system(const Poisson_system & p): ODE(p){};

        virtual valarray<double> compute_flux(valarray<double> x);
        // Pk on refini pas solve  ?car le meme solver ici 
        void set_H( double (*p)(valarray<double> a));
        void set_dH( valarray<double> (*p)(valarray<double> a));
        void set_Jx( valarray<double> (*p)(valarray<double> a,valarray<double> b));

};

template<int p> class Hamiltonian_system: public Poisson_system{
    public:
        Hamiltonian_system(): Poisson_system(){}; // justify un peu
        Hamiltonian_system(valarray<double> X0): Poisson_system(2*p,X0){};
        Hamiltonian_system(const Hamiltonian_system & h): Poisson_system(h){};

        valarray<double> compute_flux(valarray<double> x);
        void solve_dt(double );
};

template<int p> valarray<double> Hamiltonian_system<p>::compute_flux(valarray<double> x){
    valarray <double> y(n), z(n);
    y = GradH(x);
    for(int i=0;i<p;i++){
        z[i]=y[i+p];
        z[i+p]=-y[i];
    }
    // + 0.5 point coder avec des slice
    return z;
}

template<int p> void Hamiltonian_system<p>::solve_dt(double dt){
    // commenter cette fonction
    valarray<double> qpn(n);
    valarray<double> pn(p), qn(p), pn1(p),qn1(p),pstar(p), gradpn(p), gradqn(p);

    qpn = Xn.back();
    pn  = qpn[slice(p,p,1)];
    gradqn = compute_flux(qpn)[slice(0,p,1)];
    pstar = pn -0.5*dt*gradqn;

    qpn[slice(p,p,1)]= pstar;
    gradpn = compute_flux(qpn)[slice(p,p,1)];
    qn = (Xn.back())[slice(0,p,1)];
    qn1 = qn+dt*gradpn;

    qpn[slice(0,p,1)]= qn1;
    gradqn = compute_flux(qpn)[slice(0,p,1)];
    pn1 = pstar-0.5*dt*gradqn;

    qpn[slice(p,p,1)] = pn1;
    Xn.push_back(qpn);

} 

void timeloop(ODE * po, double Tf, double dt);
#endif
